@@ -7,10 +7,10 @@
 //! to another test function you must rewrite all cases or write some macros that do the job.
 //! Both solutions have some drawbreak:
 //!
-//! - rewrite create duplication
+//! - introduce duplication
 //! - macros makes code harder to read and shift out the focus from tests core
 //!
-//! The aim of this crate is solve this problem. `rstest_resuse` expose two attributes:
+//! The aim of this crate is solve this problem. `rstest_reuse` expose two attributes:
 //!
 //! - `#[template]`: to define a template
 //! - `#[apply]`: to apply a defined template to create tests
@@ -25,20 +25,16 @@
 //! // * The test list name to `two_simple_cases`
 //! // * cases: here two cases that feed the `a`, `b` values
 //! #[template]
-//! #[rstest(a,  b,
-//!     case(2, 2),
-//!     case(4/2, 2),
-//!     )
-//! ]
-//! fn two_simple_cases(a: u32, b: u32) {}
+//! #[rstest]
+//! #[case(2, 2)]
+//! #[case(4/2, 2)]
+//! fn two_simple_cases(#[case] a: u32,#[case] b: u32) {}
 //!
 //! // Here we apply the `two_simple_cases` template: That is expanded in
-//! // #[rstest(a,  b,
-//! //     case(2, 2),
-//! //     case(4/2, 2),
-//! //     )
-//! // ]
-//! // fn it_works(a: u32, b: u32) {
+//! // #[rstest]
+//! // #[case(2, 2)]
+//! // #[case(4/2, 2)]
+//! // fn it_works(#[case] a: u32,#[case] b: u32) {
 //! //     assert!(a == b);
 //! // }
 //! #[apply(two_simple_cases)]
@@ -69,12 +65,10 @@
 //! failures:
 //!
 //! ---- it_fail::case_1 stdout ----
-//! -------------- TEST START --------------
 //! thread 'it_fail::case_1' panicked at 'assertion failed: a != b', src/main.rs:34:5
 //! note: run with `RUST_BACKTRACE=1` environment variable to display a backtrace
 //!
 //! ---- it_fail::case_2 stdout ----
-//! -------------- TEST START --------------
 //! thread 'it_fail::case_2' panicked at 'assertion failed: a != b', src/main.rs:34:5
 //!
 //!
@@ -89,18 +83,96 @@
 //!
 //! Simple and neat!
 //!
+//! Note that if the test arguments names match the template's ones you can don't
+//! repeate the arguments attributes.
+//!
+//! ## Composition and Values
+//!
+//! If you need to add some cases or values when apply a template you can leverage on
+//! composition. Here a simple example:
+//!
+//! ```
+//! use rstest::rstest;
+//! use rstest_reuse::{self, *};
+//!
+//! #[template]
+//! #[rstest]
+//! #[case(2, 2)]
+//! #[case(4/2, 2)]
+//! fn base(#[case] a: u32, #[case] b: u32) {}
+//!
+//! // Here we add a new case and an argument in a value list:
+//! #[apply(base)]
+//! #[case(9/3, 3)]
+//! fn it_works(a: u32, b: u32, #[values("a", "b")] t: &str) {
+//!     assert!(a == b);
+//!     assert!("abcd".contains(t))
+//! }
+//! ```
+//!
+//! `cargo test` runs 6 tests:
+//!
+//! ```text
+//! running 6 tests
+//! test it_works::case_1::t_2 ... ok
+//! test it_works::case_2::t_2 ... ok
+//! test it_works::case_2::t_1 ... ok
+//! test it_works::case_3::t_2 ... ok
+//! test it_works::case_3::t_1 ... ok
+//! test it_works::case_1::t_1 ... ok
+//! ```
+//!
+//! Template can also used for `#[values]` and `#[with]` arguments if you need:
+//!
+//! ```
+//! use rstest::*;
+//! use rstest_reuse::{self, *};
+//!
+//! #[template]
+//! #[rstest]
+//! fn base(#[with(42)] fix: u32, #[values(1,2,3)] v: u32) {}
+//!
+//! #[fixture]
+//! fn fix(#[default(0)] inner: u32) -> u32 {
+//!     inner
+//! }
+//!
+//! #[apply(base)]
+//! fn use_it_with_fixture(fix: u32, v: u32) {
+//!     assert!(fix%v == 0);
+//! }
+//!
+//! #[apply(base)]
+//! fn use_it_without_fixture(v: u32) {
+//!     assert!(24 % v == 0);
+//! }
+//! ```
+//!
+//! `cargo test` runs 6 tests:
+//!
+//! ```text
+//! running 6 tests
+//! test use_it_with_fixture::v_1 ... ok
+//! test use_it_without_fixture::v_1 ... ok
+//! test use_it_with_fixture::v_3 ... ok
+//! test use_it_without_fixture::v_2 ... ok
+//! test use_it_without_fixture::v_3 ... ok
+//! test use_it_with_fixture::v_2 ... ok
+//! ```
+//!
+//!
 //! ## Cavelets
 //!
-//! ### `use rstest_resuse` at the top of your crate
+//! ### `use rstest_reuse` at the top of your crate
 //!
-//! You **should** add `use rstest_resuse` at the top of your crate:
+//! You **should** add `use rstest_reuse` at the top of your crate:
 //!
 //! ```
 //! #[cfg(test)]
 //! use rstest_reuse;
 //! ```
 //!
-//! This is due `rstest_reuse::template` define a macro that need to call a `rstest_resuse`'s macro.
+//! This is due `rstest_reuse::template` define a macro that need to call a `rstest_reuse`'s macro.
 //! I hope to remove this in the future but for now we should live with it.
 //!
 //! Note that
@@ -109,26 +181,6 @@
 //! ```
 //! is not enougth: this statment doesn't include `rstest_reuse` but just its public items.
 //!
-//!
-//! ### Define `template` before `apply` it
-//!
-//! `template` attribute define a macro that `apply` will use. Macro in rust are expanded in
-//! a single depth-first, lexical-order traversal of a crate’s source, that means the template
-//! definition should be allways before the `apply`.
-//!
-//! ### Tag modules with `#[macro_use]`
-//!
-//! If you define a `template` in a module and you want to use it outside the module you should
-//! _lift_ it by mark the module with the `#[macro_use]` attribute. This attribute make your
-//! `template` visibe outside this module but not at the upper level. When a `template` is
-//! defined you can use it in all submodules that follow the definition.
-//!
-//! If you plan to spread your templates in some modules and you use a unique name for each template
-//! consider to add the global attribute `!#[macro_use]` at crate level: this put all your templates
-//! available everywhere: you should
-//! just take care that a `template` should be defined before the `apply` call.
-//!
-//!
 //! ## Disclamer
 //!
 //! This crate is in developer stage. I don't know if I'll include it in `rstest` or changing some syntax in
@@ -138,9 +190,13 @@
 //! [`rstest`](https://github.com/la10736/rstest)
 
 extern crate proc_macro;
+use std::collections::HashMap;
+
 use proc_macro::TokenStream;
 use quote::{format_ident, quote};
-use syn::{self, parse, parse::Parse, parse_macro_input, Attribute, ItemFn, Path, Token};
+use syn::{
+    self, parse, parse::Parse, parse_macro_input, Attribute, Ident, ItemFn, PatType, Path, Token,
+};
 
 struct MergeAttrs {
     template: ItemFn,
@@ -174,6 +230,56 @@ fn sanitize_should_panic_duplication_bug(
     attributes
 }
 
+fn collect_template_args(template: &ItemFn) -> HashMap<&Ident, &PatType> {
+    template
+        .sig
+        .inputs
+        .iter()
+        .filter_map(|arg| match arg {
+            syn::FnArg::Typed(a) => Some(a),
+            _ => None,
+        })
+        .filter_map(|arg| match *arg.pat {
+            syn::Pat::Ident(ref id) => Some((&id.ident, arg)),
+            _ => None,
+        })
+        .collect()
+}
+
+fn merge_arg_attributes(dest: &mut Vec<Attribute>, source: &[Attribute]) {
+    for s in source.iter() {
+        if !dest.contains(s) {
+            dest.push(s.clone())
+        }
+    }
+}
+
+fn resolve_template_arg<'a>(
+    template: &HashMap<&'a Ident, &'a PatType>,
+    arg: &Ident,
+) -> Option<&'a PatType> {
+    let id_name = arg.to_string();
+    match (template.get(arg), id_name.starts_with('_')) {
+        (Some(&arg), _) => Some(arg),
+        (None, true) => template.get(&format_ident!("{}", id_name[1..])).copied(),
+        _ => None,
+    }
+}
+
+fn expand_function_arguments(dest: &mut ItemFn, source: &ItemFn) {
+    let to_merge_args = collect_template_args(source);
+
+    for arg in dest.sig.inputs.iter_mut() {
+        if let syn::FnArg::Typed(a) = arg {
+            if let syn::Pat::Ident(ref id) = *a.pat {
+                if let Some(source_arg) = resolve_template_arg(&to_merge_args, &id.ident) {
+                    merge_arg_attributes(&mut a.attrs, &source_arg.attrs);
+                }
+            }
+        }
+    }
+}
+
 #[doc(hidden)]
 #[proc_macro]
 pub fn merge_attrs(item: TokenStream) -> TokenStream {
@@ -181,8 +287,10 @@ pub fn merge_attrs(item: TokenStream) -> TokenStream {
         template,
         mut function,
     } = parse_macro_input!(item as MergeAttrs);
-    let mut attrs = template.attrs;
 
+    expand_function_arguments(&mut function, &template);
+
+    let mut attrs = template.attrs;
     #[cfg(sanitize_multiple_should_panic_compiler_bug)]
     {
         function.attrs = sanitize_should_panic_duplication_bug(function.attrs);
@@ -198,8 +306,8 @@ pub fn merge_attrs(item: TokenStream) -> TokenStream {
 
 fn get_export(attributes: &[Attribute]) -> Option<&Attribute> {
     attributes
-        .into_iter()
-        .find(|&attr| attr.path.is_ident(&format_ident!("export")))
+        .iter()
+        .find(|&attr| attr.path().is_ident(&format_ident!("export")))
 }
 
 /// Define a template where the name is given from the function name. This attribute register all
@@ -207,9 +315,12 @@ fn get_export(attributes: &[Attribute]) -> Option<&Attribute> {
 /// use a signature like if you're wrinting a standard `rstest`.
 ///
 /// If you need to export the template at the root of your crate or use it from another crate you
-/// should annotate it with `#[export]` attribute. This attribute add #[macro_export] attribute to 
+/// should annotate it with `#[export]` attribute. This attribute add `#[macro_export]` attribute to
 /// the template macro and make possible to use it from another crate.
 ///
+/// When define a template you can also set the arguments attributes like `#[case]`, `#[values]`
+/// and `#[with]`: when you apply it attributes will be copied to the matched by name arguments.
+///
 #[proc_macro_attribute]
 pub fn template(_args: proc_macro::TokenStream, input: proc_macro::TokenStream) -> TokenStream {
     let mut template: ItemFn = parse(input).unwrap();
@@ -217,7 +328,7 @@ pub fn template(_args: proc_macro::TokenStream, input: proc_macro::TokenStream)
     let rstest_index = template
         .attrs
         .iter()
-        .position(|attr| attr.path.is_ident(&format_ident!("rstest")));
+        .position(|attr| attr.path().is_ident(&format_ident!("rstest")));
 
     let mut attributes = template.attrs;
 
@@ -226,19 +337,30 @@ pub fn template(_args: proc_macro::TokenStream, input: proc_macro::TokenStream)
         None => std::mem::take(&mut attributes),
     };
 
-    let mut tokens = match get_export(&attributes) {
-        Some(_) => {
+    let (macro_attribute, visibility) = match get_export(&attributes) {
+        Some(_) => (
             quote! {
                 #[macro_export]
-            }
-        }
-        None => quote! {},
+            },
+            quote! {
+                pub
+            },
+        ),
+        None => (
+            quote! {},
+            quote! {
+                pub(crate)
+            },
+        ),
     };
 
     let macro_name = template.sig.ident.clone();
-    tokens.extend(quote! {
+    let macro_name_rand = format_ident!("{}_{}", macro_name, rand::random::<u64>());
+
+    let tokens = quote! {
         /// Apply #macro_name template to given body
-        macro_rules! #macro_name {
+        #macro_attribute
+        macro_rules! #macro_name_rand {
             ( $test:item ) => {
                         $crate::rstest_reuse::merge_attrs! {
                             #template,
@@ -246,7 +368,9 @@ pub fn template(_args: proc_macro::TokenStream, input: proc_macro::TokenStream)
                         }
                     }
         }
-    });
+        #[allow(unused_imports)]
+        #visibility use #macro_name_rand as #macro_name;
+    };
     tokens.into()
 }
 
@@ -264,12 +388,10 @@ pub fn template(_args: proc_macro::TokenStream, input: proc_macro::TokenStream)
 /// }
 ///
 /// #[template]
-/// #[rstest(a,  b,
-///     case(2, 2),
-///     case(4/2, 2),
-///     )
-/// ]
-/// fn two_simple_cases(a: u32, b: u32) {}
+/// #[rstest]
+/// #[case(2, 2)]
+/// #[case(4/2, 2)]
+/// fn two_simple_cases(#[case] a: u32, #[case] b: u32) {}
 ///
 /// #[apply(two_simple_cases)]
 /// fn it_works(mut empty: Vec<u32>, a: u32, b: u32) {
@@ -277,6 +399,48 @@ pub fn template(_args: proc_macro::TokenStream, input: proc_macro::TokenStream)
 ///     assert!(empty.last() == b);
 /// }
 /// ```
+/// When use `#[apply]` you can also
+/// 1. Use a path for template
+/// 2. Ignore an argument by underscore
+/// 3. add some cases
+/// 4. add some values
+///
+///
+/// ```
+/// use rstest::{rstest, fixture};
+/// use rstest_reuse::{self, *};
+///
+/// #[fixture]
+/// fn fix (#[default(0)] inner: u32) -> u32 {
+///     inner
+/// }
+///
+/// mod outer {
+///     pub(crate) mod inner {
+///         use rstest_reuse::template;
+///
+///         #[template]
+///         #[rstest]
+///         #[case(2, 2)]
+///         #[case(4/2, 2)]
+///         fn two_simple_cases(#[case] a: u32, #[case] b: u32) {}
+///     }
+/// }
+///
+///
+/// #[apply(outer::inner::two_simple_cases)]
+/// // Add a case
+/// #[case(9/3, 3)]
+/// // Use fixture with 42 as argument
+/// // Ignore b case values
+/// // add 2 cases with other in 4, 5 for each case
+/// fn lot_of_tests(fix: u32, a: u32, _b: u32, #[values(4, 5)] other: u32) {
+///     assert_eq!(fix, 42);
+///     assert_eq!(a, 2);
+///     assert!([4, 5].contains(other));
+/// }
+/// ```
+///
 
 #[proc_macro_attribute]
 pub fn apply(args: proc_macro::TokenStream, input: proc_macro::TokenStream) -> TokenStream {
